from inspect import iscoroutinefunction
from typing import Awaitable, Callable, Protocol

from promplate.llm.base import *
from promplate.prompt import Template
from promplate.prompt.template import AutoNaming, Context

from .utils import appender, count_position_parameters

PostProcessReturns = Context | None | tuple[Context | None, str]

PreProcess = Callable[[Context], Context]
PostProcess = (
    Callable[[Context], PostProcessReturns]
    | Callable[[Context, str], PostProcessReturns]
)

AsyncPreProcess = Callable[[Context], Awaitable[Context]]
AsyncPostProcess = (
    Callable[[Context], Awaitable[PostProcessReturns]]
    | Callable[[Context, str], Awaitable[PostProcessReturns]]
)


class AbstractChain(Protocol):
    def run(
        self,
        context: Context,
        complete: Complete | None = None,
    ) -> Context:
        ...

    async def arun(
        self,
        context: Context,
        complete: Complete | AsyncComplete | None = None,
    ) -> Context:
        ...

    complete: Complete | AsyncComplete | None


class Interruptable(AbstractChain, Protocol):
    def _run(
        self,
        context: Context,
        complete: Complete | None = None,
    ) -> Context:
        ...

    async def _arun(
        self,
        context: Context,
        complete: Complete | AsyncComplete | None = None,
    ) -> Context:
        ...

    def run(self, context, complete=None) -> Context:
        complete = complete or self.complete
        try:
            return self._run(context, complete)
        except JumpTo as jump:
            if jump.target is None or jump.target is self:
                return jump.chain.run(jump.context or context, complete)
            else:
                raise jump from None

    async def arun(self, context, complete=None) -> Context:
        complete = complete or self.complete
        try:
            return await self._arun(context, complete)
        except JumpTo as jump:
            if jump.target is None or jump.target is self:
                return await jump.chain.arun(jump.context or context, complete)
            else:
                raise jump from None


class Node(AutoNaming, Interruptable):
    def __init__(
        self,
        template: Template,
        pre_processes: list[PreProcess | AsyncPreProcess] | None = None,
        post_processes: list[PostProcess | AsyncPostProcess] | None = None,
        complete: Complete | AsyncComplete | None = None,
        **config,
    ):
        self.template = template
        self.pre_processes = pre_processes or []
        self.post_processes = post_processes or []
        self.complete = complete
        self.run_config = config

    @property
    def pre_process(self):
        return appender(self.pre_processes)

    @property
    def post_process(self):
        return appender(self.post_processes)

    @staticmethod
    def _via(process: PostProcess | AsyncPostProcess, context: Context, result):
        if count_position_parameters(process) == 1:
            return process(context)
        else:
            return process(context, result)

    def _apply_pre_processes(self, context: Context):
        for process in self.pre_processes:
            context = process(context) or context
        return context

    def _apply_post_processes(self, context: Context, result):
        for process in self.post_processes:
            ret = self._via(process, context, result) or context
            if isinstance(ret, tuple):
                ret, result = ret
            context = ret or context

        return context

    def _run(self, context, complete=None):
        complete = complete or self.complete
        assert complete is not None

        context = self._apply_pre_processes(context)
        prompt = self.template.render(context)

        result = context["__result__"] = complete(prompt, **self.run_config)

        context = self._apply_post_processes(context, result)

        return context

    async def _apply_async_pre_processes(self, context: Context):
        for process in self.pre_processes:
            if iscoroutinefunction(process):
                context = await process(context) or context
            else:
                context = process(context) or context

        return context

    async def _apply_async_post_processes(self, context: Context, result):
        for process in self.post_processes:
            if iscoroutinefunction(process):
                ret = await self._via(process, context, result)
            else:
                ret = self._via(process, context, result)

            if isinstance(ret, tuple):
                ret, result = ret
            context = ret or context

        return context

    async def _arun(self, context, complete=None):
        complete = complete or self.complete
        assert complete is not None

        context = await self._apply_async_pre_processes(context)
        prompt = await self.template.arender(context)

        if iscoroutinefunction(complete):
            result = context["__result__"] = await complete(prompt, **self.run_config)
        else:
            result = context["__result__"] = complete(prompt, **self.run_config)

        context = await self._apply_async_post_processes(context, result)

        return context

    def next(self, chain: AbstractChain):
        if isinstance(chain, Chain):
            return Chain(self, *chain)
        else:
            return Chain(self, chain)

    def __add__(self, chain: AbstractChain):
        return self.next(chain)

    def render(self, context: Context):
        context = self._apply_pre_processes(context)
        return self.template.render(context)

    async def arender(self, context: Context):
        context = await self._apply_async_pre_processes(context)
        return await self.template.arender(context)

    def __str__(self):
        return f"</{self.name}/>"


class Chain(Interruptable):
    def __init__(
        self,
        *nodes: AbstractChain,
        complete: Complete | AsyncComplete | None = None,
    ):
        self.nodes = list(nodes)
        self.complete = complete

    def next(self, chain: AbstractChain):
        if isinstance(chain, Node):
            return Chain(*self, chain)
        elif isinstance(chain, Chain):
            return Chain(*self, *chain)
        else:
            raise NotImplementedError

    def __add__(self, chain):
        return self.next(chain)

    def __iter__(self):
        return iter(self.nodes)

    def _run(self, context, complete=None):
        for node in self.nodes:
            context = node.run(context, node.complete or complete)

        return context

    async def _arun(self, context, complete=None):
        for node in self.nodes:
            context = await node.arun(context, node.complete or complete)

        return context

    def __repr__(self):
        return " + ".join(map(str, self.nodes))


class JumpTo(Exception):
    def __init__(
        self,
        chain: Interruptable,
        context: Context | None = None,
        target: Interruptable | None = None,
    ):
        self.chain = chain
        self.context = context
        self.target = target

    def __str__(self) -> str:
        return f"{self.target} does not exist in the hierarchy"



from promplate.prompt.chat import Message, parse_chat_markup


def ensure(text_or_list: list[Message] | str) -> list[Message]:
    return (
        parse_chat_markup(text_or_list)
        if isinstance(text_or_list, str)
        else text_or_list
    )



from inspect import Parameter, signature


def appender(to_append: list):
    def append_processer(func):
        to_append.append(func)

        return func

    return append_processer


def is_positional_parameter(p: Parameter):
    return p.kind is Parameter.POSITIONAL_OR_KEYWORD or p.kind is Parameter.KEYWORD_ONLY


def count_position_parameters(func):
    return sum(map(is_positional_parameter, signature(func).parameters.values()))



from copy import copy
from functools import cached_property
from pathlib import Path
from typing import Any, Protocol

from .builder import *
from .utils import *

Context = dict[str, Any]  # globals must be a real dict


class Component(Protocol):
    def render(self, context: Context) -> str:
        ...

    async def arender(self, context: Context) -> str:
        ...


class TemplateCore(AutoNaming):
    """A simple template compiler, for a jinja2-like syntax."""

    def __init__(self, text: str):
        """Construct a Templite with the given `text`."""

        self.text = text
        self._buffer = []
        self._ops_stack = []

    def _flush(self):
        if len(self._buffer) == 1:
            self._builder.add_line(f"append_result({self._buffer[0]})")
        elif len(self._buffer) > 1:
            self._builder.add_line(f"extend_result(({', '.join(self._buffer)}))")
        self._buffer.clear()

    @staticmethod
    def _unwrap_token(token: str):
        return token.strip()[2:-2].strip("-").strip()

    def _on_literal_token(self, token: str):
        self._buffer.append(repr(token))

    def _on_eval_token(self, token):
        exp = self._unwrap_token(token)
        self._buffer.append(f"str({exp})")

    def _on_comment_token(self, _):
        pass  # TODO: remove blank line

    def _on_special_token(self, token, sync: bool):
        inner: str = self._unwrap_token(token)

        if inner.startswith("end"):
            last = self._ops_stack.pop()
            assert last == inner.removeprefix("end")
            self._flush()
            self._builder.dedent()

        else:
            op = inner.split(" ", 1)[0]

            if op == "if" or op == "for":
                self._ops_stack.append(op)
                self._flush()
                self._builder.add_line(f"{inner}:")
                self._builder.indent()

            elif op == "else" or op == "elif":
                self._flush()
                self._builder.dedent()
                self._builder.add_line(f"{inner}:")
                self._builder.indent()

            else:
                arguments: str = self._make_context(inner)
                if sync:
                    self._buffer.append(f"{op}.render({arguments})")
                else:
                    self._buffer.append(f"await {op}.arender({arguments})")

    @staticmethod
    def _make_context(text: str):
        pairs = []
        for token in text.split(" ")[1:]:
            if token == "*":
                pairs.append("**locals()")
            elif "=" in token or token.startswith("**"):
                pairs.append(token)
            else:
                pairs.append(f"{token}={token}")

        return f"dict({', '.join(pairs)})"

    def compile(self, sync=True):
        self._builder = get_base_builder(sync)

        for token in split_template_tokens(self.text):
            s_token = token.strip()
            if not s_token.startswith("{"):
                self._on_literal_token(token)
                continue
            if s_token.startswith("{{"):
                self._on_eval_token(token)
            elif s_token.startswith("{%"):
                self._on_special_token(token, sync)
            else:
                assert s_token.startswith("{#")
                self._on_comment_token(token)

        if self._ops_stack:
            raise SyntaxError(self._ops_stack)

        self._flush()
        self._builder.add_line("return ''.join(result)")
        self._builder.dedent()

    @cached_property
    def _render_code(self):
        self.compile()
        return self._builder.get_render_function().__code__

    def render(self, context: Context) -> str:
        return eval(self._render_code, copy(context))

    @cached_property
    def _arender_code(self):
        self.compile(sync=False)
        return self._builder.get_render_function().__code__

    async def arender(self, context: Context) -> str:
        return await eval(self._arender_code, copy(context))


class Template(TemplateCore):
    @classmethod
    def read(cls, path: str | Path, encoding="utf-8"):
        path = Path(path)
        template = cls(path.read_text(encoding))
        template.name = path.stem
        return template

    @classmethod
    async def aread(cls, path: str | Path, encoding="utf-8"):
        from aiofiles import open

        async with open(path, encoding=encoding) as f:
            content = await f.read()

        path = Path(path)
        template = cls(content)
        template.name = path.stem
        return template

    _client = None

    @classmethod
    def fetch(cls, url: str, **kwargs):
        if cls._client is None:
            from httpx import Client

            cls._client = Client(**kwargs)

        response = cls._client.get(url)
        template = cls(response.text)
        template.name = Path(url).stem
        return template

    _aclient = None

    @classmethod
    async def afetch(cls, url: str, **kwargs):
        if cls._aclient is None:
            from httpx import AsyncClient

            cls._aclient = AsyncClient(**kwargs)

        response = await cls._aclient.get(url)
        template = cls(response.text)
        template.name = Path(url).stem
        return template



from typing import Literal, TypedDict

from .template import Template
from .utils import is_message_start

Message = TypedDict(
    "Message",
    {"role": Literal["user", "assistant", "system"], "content": str, "name": str},
    total=False,
)


def parse_chat_markup(text: str) -> list[Message]:
    messages = []
    current_message = None
    buffer = []

    for line in text.splitlines():
        match = is_message_start.match(line)
        if match:
            role, name = match.group(1), match.group(2)

            if current_message:
                current_message["content"] = "\n".join(buffer)
                messages.append(current_message)
                buffer.clear()

            current_message = {"role": role, "content": ""}

            if name:
                current_message["name"] = name

        elif current_message:
            buffer.append(line)

    if current_message:
        current_message["content"] = "\n".join(buffer)
        messages.append(current_message)

    return messages or [{"role": "user", "content": text}]


class ChatTemplate(Template):
    def render(self, context) -> list[Message]:
        return parse_chat_markup(super().render(context))

    async def arender(self, context) -> list[Message]:
        return parse_chat_markup(await super().arender(context))



from typing import *

from promplate.prompt import Message

from . import base

class Configurable(base.Configurable):
    def __init__(
        self,
        *,
        model: str,
        temperature: float | int | None = None,
        top_p: float | int | None = None,
        stop: str | list[str] | None = None,
        max_tokens: int | None = None,
        api_key: str | None = None,
        api_base: str | None = None,
        **other_config,
    ):
        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.stop = stop
        self.max_tokens = max_tokens
        self.api_key = api_key
        self.api_base = api_base

class TextComplete(Configurable, base.Complete):
    def __call__(
        self,
        text: str,
        /,
        **config,
    ) -> str: ...

class AsyncTextComplete(Configurable, base.AsyncComplete):
    def __call__(
        self,
        text: str,
        /,
        **config,
    ) -> Awaitable[str]: ...

class TextGenerate(Configurable, base.Generate):
    def __call__(
        self,
        text: str,
        /,
        **config,
    ) -> Iterable[str]: ...

class AsyncTextGenerate(Configurable, base.AsyncGenerate):
    def __call__(
        self,
        text: str,
        /,
        **config,
    ) -> AsyncIterable[str]: ...

class ChatComplete(Configurable, base.Complete):
    def __call__(
        self,
        messages: list[Message] | str,
        /,
        **config,
    ) -> str: ...

class AsyncChatComplete(Configurable, base.AsyncComplete):
    def __call__(
        self,
        messages: list[Message] | str,
        /,
        **config,
    ) -> Awaitable[str]: ...

class ChatGenerate(Configurable, base.Generate):
    def __call__(
        self,
        messages: list[Message] | str,
        /,
        **config,
    ) -> Iterable[str]: ...

class AsyncChatGenerate(Configurable, base.AsyncGenerate):
    def __call__(
        self,
        messages: list[Message] | str,
        /,
        **config,
    ) -> AsyncIterable[str]: ...

def text_complete(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> TextComplete: ...
def async_text_complete(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> AsyncTextComplete: ...
def text_generate(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> TextGenerate: ...
def async_text_generate(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> AsyncTextGenerate: ...
def chat_complete(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> ChatComplete: ...
def async_chat_complete(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> AsyncChatComplete: ...
def chat_generate(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> ChatGenerate: ...
def async_chat_generate(
    *,
    model: str,
    temperature: float | int | None = None,
    top_p: float | int | None = None,
    stop: str | list[str] | None = None,
    max_tokens: int | None = None,
    api_key: str | None = None,
    api_base: str | None = None,
    **other_config,
) -> AsyncChatGenerate: ...



from importlib.metadata import metadata

import openai
from openai import ChatCompletion, Completion

from .base import *
from .utils import Message, ensure

meta = metadata("promplate")

openai.app_info = {
    "name": "Promplate",
    "version": meta["version"],
    "url": meta["home-page"],
}


class TextComplete(Configurable, Complete):
    def __call__(self, text, /, **config):
        config = self._config | config | {"stream": False, "prompt": text}
        result = Completion.create(**config)
        return result["choices"][0]["text"]


class AsyncTextComplete(Configurable, AsyncComplete):
    async def __call__(self, text, /, **config):
        config = self._config | config | {"stream": False, "prompt": text}
        result = await Completion.acreate(**config)
        return result["choices"][0]["text"]


class TextGenerate(Configurable, Generate):
    def __call__(self, text, /, **config):
        config = self._config | config | {"stream": True, "prompt": text}
        stream = Completion.create(**config)
        for event in stream:
            yield event["choices"][0]["text"]


class AsyncTextGenerate(Configurable, AsyncGenerate):
    async def __call__(self, text, /, **config):
        config = self._config | config | {"stream": True, "prompt": text}
        stream = await Completion.acreate(**config)
        async for event in stream:
            yield event["choices"][0]["text"]


class ChatComplete(Configurable, Complete):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": False, "messages": messages}
        result = ChatCompletion.create(**config)
        return result["choices"][0]["message"]["content"]


class AsyncChatComplete(Configurable, AsyncComplete):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": False, "messages": messages}
        result = await ChatCompletion.acreate(**config)
        return result["choices"][0]["message"]["content"]


class ChatGenerate(Configurable, Generate):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": True, "messages": messages}
        stream = ChatCompletion.create(**config)
        for event in stream:
            delta: dict = event["choices"][0]["delta"]
            yield delta.get("content", "")


class AsyncChatGenerate(Configurable, AsyncGenerate):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": True, "messages": messages}
        stream = await ChatCompletion.acreate(**config)
        async for event in stream:
            delta: dict = event["choices"][0]["delta"]
            yield delta.get("content", "")


def text_complete(**default_config):
    return TextComplete(**default_config)


def async_text_complete(**default_config):
    return AsyncTextComplete(**default_config)


def text_generate(**default_config):
    return TextGenerate(**default_config)


def async_text_generate(**default_config):
    return AsyncTextGenerate(**default_config)


def chat_complete(**default_config):
    return ChatComplete(**default_config)


def async_chat_complete(**default_config):
    return AsyncChatComplete(**default_config)


def chat_generate(**default_config):
    return ChatGenerate(**default_config)


def async_chat_generate(**default_config):
    return AsyncChatGenerate(**default_config)
